(** Make a pure data structure module buildable from imperative events. *)

type id
val id_gen : unit -> (unit -> id)
(* Returns an effectful id generator:
     let id_gen = id_gen () in
     id_gen (); id_gen (); id_gen ()
   generates 3 unique IDs. *)

(** Input signature for the module to be used from imperative changes.
    It implements the data structure that will be usable through FRP.  *)
module type M = sig
  type t
  (** The type of the data structure. *)
  type init
  (** The type of the argument passed to the M.create function. *)
  type action_functional
  (** Actions indicate how to change the data structure: for instance, "`Add 42"
      to add the int 42 to a set data structure.
      action_functional is for actions to do in the functional side of the
      program (triggered from the UI). *)
  type action_ui
  (** See action_functional. action_ui is for actions to do in the UI (triggered
      from the functional side of the program). *)
  val create : init -> t * action_ui
  (** The initial state the data structure will be in. *)
  val on_action_func : t -> action_functional -> t * action_ui
  (** on_action is a function passed to a "fold" operation: from the current
   state and an action, it creates a new state.
   Here, 'a is an identifier for the source of the action. It is opaque but can
   be used to find whether the source of the action was the same as the source
   of a previous action (which is why it is stored in the datastructure). *)
end

(** Input signature for the FRP support module.
 The signature is rather light and any FRP implementation should be usable. *)
module type FRP = sig
  type 'a t
  val create : unit -> 'a t * ('a -> unit)
  (** Create a new FRP event. *)
  val fold : ('a -> 'b -> 'a) -> 'a -> 'b t -> 'a t
  (** fold over FRP events. *)
  val map : ('a -> 'b) -> 'a t -> 'b t
  (** map over FRP events. *)
end

class ['a] callbacks : ?callbacks:(('a -> unit) list) -> unit -> object
  method set : ('a -> unit) list -> unit
  method get : ('a -> unit) list
  method add : ('a -> unit) -> unit
  method remove : ('a -> unit) -> unit
  method call : 'a -> unit
end

(** Makes a pure data structure usable from imperative events. *)
module Core : functor (FRP : FRP) -> functor (M : M) -> sig
  class core :
    ?state_callbacks:(M.t * M.action_ui -> unit) list ->
    ?msg_callbacks:(M.action_ui -> unit) list ->
    M.init ->
    object
      method msg_callbacks : M.action_ui callbacks
      (** Callbacks for messages from the data structure update. *)
      method send : M.action_functional -> unit
      (** Send an action in order to make the functional state change. *)
      method state_callbacks : (M.t * M.action_ui) callbacks
      (** Callbacks for changes of the data structure. *)
    end
end

